#include <assert.h>
#include <stdio.h>
#define __STDC_LIMIT_MACROS
#include <stdlib.h>
#include <stdint.h>

#include <vector>
#include <iostream>

using namespace std;

#include "fompi.h"
#include "Locks.h"

#define MAX_AVG_COLLISIONS 4
//#define HTABLE_FILL_RATIO 0.01

int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

MCSLock* mcs_lock;

int64_t bigRandVal() {

  int64_t random =
    (((int64_t) rand() <<  0) & 0x000000000000FFFFull) |
    (((int64_t) rand() << 16) & 0x00000000FFFF0000ull) |
    (((int64_t) rand() << 32) & 0x0000FFFF00000000ull) |
    (((int64_t) rand() << 48) & 0x0FFF000000000000ull);
  return random;

}

int64_t TSIZE;

int insert_collisions = 0;
int delete_collisions = 0;
int inserts = 0;
int deletes = 0;

extern "C" {
  typedef struct{
    int64_t value;
    int64_t next;
  } t_elem;
}

typedef struct {
  foMPI_Win twin; // table window
  foMPI_Win lwin; // last entry window
  foMPI_Win nwin; // nextfree window
  t_elem *table; // the table + heap
  int64_t nextfree; // the next free item in heap
  int64_t size; // size of the heap
  int64_t *last; // direct link to last element -- avoid traversing list!
  int p, r;
  int64_t tsize; // size of table
} t_hash;

int64_t hashfunc(int64_t val) {
  return val & (TSIZE-1);
}

void printHTAll(t_hash* hash, int n) {

  int val = -1;
  int nxt = -1;
  int owner = 0;
  int offset = 0;
  fprintf(stderr,"[%i] @@@@@@ Table: \n ",hash->r);
  for(int i = 0; i < TSIZE; i++) {
    if(i % hash->p == 0 && i > 0) {
      owner++;
      offset = 0;
      fprintf(stderr,"--\n");
    }

    foMPI_Get(&val,1,MPI_INT,owner,offset,1,MPI_INT,hash->twin);
    foMPI_Win_flush(owner, hash->twin);
    foMPI_Get(&nxt,1,MPI_INT,owner,offset+1,1,MPI_INT,hash->twin);
    foMPI_Win_flush(owner, hash->twin);
    fprintf(stderr,">> [%i] %i : (value: %i, next: %i)\n",owner,i,val,nxt);
    offset += 2;
  }

  owner = 0;
  offset = 0;

  fprintf(stderr,"\n[%i] @@@ Heap: \n",hash->r);
  for(int i = 0; i < n; i++) {

    if(i % hash->p == 0 && i > 0) {
      owner++;
      offset = 0;
      fprintf(stderr,"--\n");
    }

    foMPI_Get(&val,1,MPI_INT,owner,hash->tsize*2 + offset,1,MPI_INT,hash->twin);
    foMPI_Win_flush(owner, hash->twin);
    foMPI_Get(&nxt,1,MPI_INT,owner,hash->tsize*2 + offset+1,1,MPI_INT,hash->twin);
    foMPI_Win_flush(owner, hash->twin);
    fprintf(stderr,">> [%i] %i : (value: %i, next: %i)\n",owner,i+TSIZE,val,nxt);
    offset += 2;

  } 
}

void printHTLocal(t_hash* hash) {
  printf("[%i] @@@@@@ Table: \n ",hash->r);
  for(int i = 0; i < hash->tsize; i++) {    
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i].value, (hash->table)[i].next);
  }

  printf("\n[%i] @@@ Heap: \n",hash->r);
  for(int i = hash->tsize; i < hash->tsize+hash->size; i++) {
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i].value, (hash->table)[i].next); 
  }
}

int64_t get(t_hash* hash, int64_t pos) {
  int64_t owner = pos/hash->tsize;
  int64_t result = -2;
  int64_t offset /* in ints */ = (pos-owner*hash->tsize)*2;

  //foMPI_Win_lock(foMPI_LOCK_SHARED, owner, 0, hash->twin);
  //ReaderLockAcquire();
  //MCSLockAcquire();
  mcs_lock[owner].acquire();

  foMPI_Get(&result, 1, MPI_INT64_T, owner, offset, 1, MPI_INT64_T, hash->twin);
  ///foMPI_Fetch_and_op(/* origin */ NULL, /* result */ &result, MPI_INT64_T, owner, /* disp */ offset, foMPI_NO_OP, hash->twin);
  foMPI_Win_flush_local(owner, hash->twin);

  //foMPI_Win_unlock(owner, hash->twin);
  //ReaderLockRelease();
  //MCSLockRelease();
  mcs_lock[owner].release();

  return result;
}

void insert(t_hash* hash, int64_t elem) {
  int64_t pos = hashfunc(elem);
  int64_t owner = pos/hash->tsize;

  { // remote update

    //foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, owner, 0, hash->twin);
    //MCSLockAcquire();
    mcs_lock[owner].acquire();

    // CAS if table position is available
    int64_t compare = -1;
    int64_t result = -2;
    int64_t offset /* in ints */ = (pos-owner*hash->tsize)*2;    

    foMPI_Get(&result, 1, MPI_INT64_T, owner, offset, 1, MPI_INT64_T, hash->twin);
    foMPI_Win_flush_local(owner, hash->twin);

    ///foMPI_Compare_and_swap(&elem, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->twin);
    ///foMPI_Win_flush_local(owner, hash->twin);

    if(result == compare) {
      foMPI_Put(&elem, 1, MPI_INT64_T, owner, offset, 1, MPI_INT64_T, hash->twin);
      foMPI_Win_flush(owner, hash->twin);
    }
    else { // we lost the CAS, no direct add
      int64_t offset = (pos-owner*hash->tsize);
      int64_t origin = 1; // add 1
      int64_t result = -1; // address of newly added record

      // grab remote location
      foMPI_Fetch_and_op(/* origin */ &origin, /* result */ &result, MPI_INT64_T, owner, /* disp */ 0, foMPI_SUM, hash->nwin);
      foMPI_Win_flush_local(owner, hash->nwin);

      if(result >= hash->tsize+hash->size) {
        fprintf(stderr,"Failed to insert due to the OVERFLOW\n");
        return;
      }

      // put element in grabbed location
      foMPI_Put(&elem, 1, MPI_INT64_T, owner, 2*(result), 1, MPI_INT64_T, hash->twin);

      // try to put directly as table link
      int64_t newOffset = result;

      int64_t oldOffset = -111;    // mine
      offset = (pos-owner*hash->tsize);

      // grab lastptr element and plug in ptr to new element
      foMPI_Fetch_and_op(/* origin */ &newOffset, /* result */ &oldOffset, MPI_INT64_T, owner, /* disp */ offset, foMPI_REPLACE, hash->lwin);
      foMPI_Win_flush_local(owner, hash->lwin);

      compare = -1;
      result = -2;
      offset /* in ints */ = (pos-owner*hash->tsize)*2+1;

      // try to update table pointer directly (if it's the first update)
      foMPI_Compare_and_swap(&newOffset, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->twin);
      foMPI_Win_flush_local(owner, hash->twin);

      if(result != compare) { // we lost the CAS for the direct pointer -- no direct access to table!

        foMPI_Put(&newOffset, 1, MPI_INT64_T, owner, 2*(oldOffset)+1, 1, MPI_INT64_T, hash->twin);
        foMPI_Win_flush(owner, hash->twin);
      }
    }
  }

  //foMPI_Win_unlock(owner, hash->twin);
  //MCSLockRelease();
  mcs_lock[owner].release();
}

int count(t_hash *hash) {
  int cnt=0;
  int heapcnt=0;
  for(int i=0; i<hash->tsize; i++) {
    int pos = i;
    if(hash->table[pos].value != -1) {
      cnt++;
      if(hash->table[pos].next != -1) {
        heapcnt++;
        pos=hash->table[pos].next;
        while(hash->table[pos].next != -1) {
          pos = hash->table[pos].next;
          heapcnt++;
        }
      }
    }
  }

  return cnt+heapcnt;
}

int countSimple(t_hash* hash) {
  int cnt = 0;
  for(int i = 0; i < hash->tsize+hash->size;i++) {
    if(hash->table[i].value != -1) {
      cnt++;
    }
  }
  return cnt;
}


int main(int argc, char **argv) {
  foMPI_Init(&argc, &argv);

  t_hash hash;
  int64_t LV = atol(argv[1]);
  int rw_ratio = atoi(argv[2]);
  double HTABLE_FILL_RATIO = atof(argv[3]);
  int id = atoi(argv[4]);
  marker_dist = atoi(argv[5]);
  hmcs_threshold = atoi(argv[6]);
  tree_threshold = atoi(argv[7]);
  root_threshold = atoi(argv[8]);
  reader_threshold = atoi(argv[9]);
  writer_rate = atoi(argv[10]);

  MPI_Comm_size(MPI_COMM_WORLD, &hash.p);
  MPI_Comm_rank(MPI_COMM_WORLD, &hash.r);

  TSIZE = (int64_t)LV*(int64_t)hash.p;
  srand(hash.r*1000000);

  //MCSLockInit(MPI_COMM_WORLD);
  mcs_lock = new MCSLock[hash.p];
  for(int i = 0; i < hash.p; ++i) {
    mcs_lock[i].initialize();
  }

  hash.tsize = LV;
  hash.size = (MAX_AVG_COLLISIONS)*LV;

  foMPI_Win_allocate((hash.tsize+hash.size)*sizeof(t_elem), sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.table, &hash.twin);
  foMPI_Win_lock_all(0, hash.twin);

  foMPI_Win_allocate(hash.tsize*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.last, &hash.lwin);
  foMPI_Win_lock_all(0, hash.lwin);

  hash.nextfree = hash.tsize; // next free element in heap
  foMPI_Win_create(&hash.nextfree, sizeof(int64_t), 1, MPI_INFO_NULL, MPI_COMM_WORLD, &hash.nwin);
  foMPI_Win_lock_all(0, hash.nwin);

  // initialize table
  for(int i=0; i<hash.tsize+hash.size; ++i) {
    if(i<hash.tsize) hash.last[i]=-1;
    hash.table[i].next=-1;
    hash.table[i].value=-1;
  }

  double t_start = 0.0;
  double t_end = 0.0;

  int nr_of_ins = HTABLE_FILL_RATIO*LV;

  int warmups = 0.1*nr_of_ins;
  double ins_time = 0.0;

  srand(hash.r*1000000);

#ifndef HT_DEBUG
  MPI_Barrier(MPI_COMM_WORLD);

  for(int m = 0; m < nr_of_ins; m++) {
    if(m == warmups) {
      MPI_Barrier(MPI_COMM_WORLD);
      t_start = MPI_Wtime();
    }

    int rnd = rand() % 100;
    if(rnd > rw_ratio) {
      insert(&hash, bigRandVal());
    }
    else {
      get(&hash, bigRandVal() % TSIZE);
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);
  t_end = MPI_Wtime();
  ins_time = t_end - t_start;

  int size = count(&hash);
  assert(size == countSimple(&hash));

  if(hash.r == 0) {
    //printf("RMA-MCS %i %i %f %f %f %i\n", hash.p, LV, rw_ratio, HTABLE_FILL_RATIO, ins_time, id);
    printf("%i %f %i %f %f %i RMA-MCS\n", hash.p, ins_time, LV, rw_ratio, HTABLE_FILL_RATIO, id);
  }

#else //////////////////////////////////////////////////////////////

  MPI_Barrier(MPI_COMM_WORLD);

  int64_t* vals = (int64_t*)calloc(nr_of_ins,sizeof(int64_t));
  std::vector<int64_t> fetched_vals;

  MPI_Barrier(MPI_COMM_WORLD);

  if(hash.r == 0) cout << ">>> Inserting " << nr_of_ins << endl;

  int inserted = 0;

  while(inserted < nr_of_ins) {
    int64_t val = bigRandVal();
    int rnd = rand() % 100;
    if(rnd > rw_ratio) {
      insert(&hash, val);
      vals[inserted++] = val;
    }
    else {
      int64_t elem = get(&hash, bigRandVal() % TSIZE);
      fetched_vals.push_back(elem);
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if(hash.r == 0) cout << ">>> Inserts done" << endl;

  int size = count(&hash);
  assert(size == countSimple(&hash));
  //printf("[%i] cnt: %i, simpleCnt:%i\n",hash.r,size,countSimple(&hash));

  if(hash.r == 0) cout << ">>> TEST 1 done" << endl;

  MPI_Barrier(MPI_COMM_WORLD);

  int sum = -1;
  MPI_Reduce(&size, &sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

  if(hash.r == 0) {
    assert(nr_of_ins * hash.p == sum);
    printf("The sum: %i\n", sum);
  }

  if(hash.r == 0) cout << ">>> TEST 2 done" << endl;

  int64_t* elems = (int64_t*)calloc(size,sizeof(int64_t));
  int b = 0;
  for(int i = 0; i < hash.size+hash.tsize;i++) {
    if(hash.table[i].value != -1) {
      elems[b] = hash.table[i].value;
      b++;
    }
  }

  int64_t* generated[hash.p];
  int64_t* received[hash.p];
  //int64_t** generated = new int64_t*[hash.p];
  //int64_t** received = new int64_t*[hash.p];
  int lengths[hash.p];

  if(hash.r != 0) {
    //MPI_Send(elems,size,MPI_INT64_T,0,0,MPI_COMM_WORLD);
    MPI_Send(vals,nr_of_ins,MPI_INT64_T,0,0,MPI_COMM_WORLD);
  }
  else {

    generated[hash.r] = vals;

    int procs_done = 1;
    int flag = false;
    MPI_Status s;

    while(procs_done < hash.p) {
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &s);

      if(flag == true) {
        flag = false;
        int sender = s.MPI_SOURCE;
        int tag = s.MPI_TAG;
        int len = -1;
        MPI_Get_count(&s, MPI_INT64_T, &len);
        int64_t* data = (int64_t*)calloc(len,sizeof(int64_t));
        MPI_Recv(data, len, MPI_INT64_T, sender, tag, MPI_COMM_WORLD, &s);
        procs_done++;
        generated[sender] = data;
        assert(len == nr_of_ins);
      }
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if(hash.r == 0) cout << ">>> TEST 3 done" << endl;

// READING TEST

int64_t* generated_get[hash.p];
  int64_t* received_get[hash.p];
  int lengths_get[hash.p];

  if(hash.r != 0) {
    MPI_Send(&( (fetched_vals.data())[0] ) ,fetched_vals.size(),MPI_INT64_T,0,0,MPI_COMM_WORLD);
  }
  else {

    generated_get[hash.r] = &( (fetched_vals.data())[0] );
    lengths_get[hash.r] = fetched_vals.size();

    int procs_done = 1;
    int flag = false;
    MPI_Status s;

    while(procs_done < hash.p) {
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &s);

      if(flag == true) {
        flag = false;
        int sender = s.MPI_SOURCE;
        int tag = s.MPI_TAG;
        int len = -1;
        MPI_Get_count(&s, MPI_INT64_T, &len);
        int64_t* data = (int64_t*)calloc(len,sizeof(int64_t));
        MPI_Recv(data, len, MPI_INT64_T, sender, tag, MPI_COMM_WORLD, &s);
        procs_done++;
        generated_get[sender] = data;
        lengths_get[sender] = len;
      }
    }

    for(int proc = 0; proc < hash.p; ++proc) {
      for(int64_t i = 0; i < lengths_get[proc]; ++i) {
        int64_t considered_fetched_elem = generated_get[proc][i];
        bool seen = false;
        
        for(int r = 0; r < hash.p; ++r) {
          for(int64_t j = 0; j < nr_of_ins; ++j) {
            if(considered_fetched_elem == generated[r][j]) {
              seen = true;
              goto exit_loop;
            }
          }
        }
        exit_loop:
        assert(considered_fetched_elem == -1 || seen == true);
      }
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);


  if(hash.r == 0) cout << ">>> TEST READING done" << endl;





  if(hash.r != 0) {
    MPI_Send(elems,size,MPI_INT64_T,0,0,MPI_COMM_WORLD);
  }
  else {
    received[hash.r] = elems;
    lengths[hash.r] = size;

    int procs_done = 1;
    int flag = false;
    MPI_Status s;

    while(procs_done < hash.p) {
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &s);

      if(flag == true) {
        flag = false;
        int sender = s.MPI_SOURCE;
        int tag = s.MPI_TAG;
        int len = -1;
        MPI_Get_count(&s, MPI_INT64_T, &len);
        int64_t* data = (int64_t*)calloc(len,sizeof(int64_t));
        MPI_Recv(data, len, MPI_INT64_T, sender, tag, MPI_COMM_WORLD, &s);
        procs_done++;
        received[sender] = data;
        lengths[sender] = len;
      }
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);
  if(hash.r == 0) {
    int sum_of_generated = hash.p*nr_of_ins;
    int sum_of_received = 0;

    for(int i = 0; i < hash.p; i++) {
      sum_of_received += lengths[i];
    }

    assert(sum_of_generated == sum_of_received);

    int matching = 0;
    for(int i = 0; i < hash.p; i++) {
      for(int k = 0; k < nr_of_ins; k++) {

        int64_t investigated = generated[i][k];
        //printf("\n\n>>>> looking for.. %llu\n\n", generated[i][k]);
        //int found = -1;

        for(int m = 0; m < hash.p; m++) {
          for(int n = 0; n < lengths[m]; n++) {
            //printf("%llu --- ",received[m][n]);
            if(received[m][n] == investigated) {
              received[m][n] = -1;
              matching++;
              //printf("found %i\n",matching);
              goto next_check;
            }
          }
        }
        assert(1 == 0);
next_check:
;
      }
    }
    assert(matching == sum_of_received);
  }

  if(hash.r == 0) cout << ">>> TEST 4 done" << endl;

#endif

  foMPI_Win_unlock_all(hash.nwin);
  foMPI_Win_free(&hash.nwin);

  foMPI_Win_unlock_all(hash.lwin);
  foMPI_Win_free(&hash.lwin);

  foMPI_Win_unlock_all(hash.twin);
  foMPI_Win_free(&hash.twin);

  //MCSLockFinalize();
  //mcs_lock.release();
  for(int i = 0; i < hash.p; ++i) {
    mcs_lock[i].finalize();
  }

  delete [] mcs_lock;

  foMPI_Finalize();
}
