#include <stdio.h>
#include <stdlib.h>
#include "../MersenneTwister.h"
#include <mpi.h>
#include "common.h"
#include <assert.h>

int64_t TSIZE;

//#define VALIDATION_HTOR_HASHTABLE
//#define DEBUG_ON_JYC

int insert_collisions = 0;
int delete_collisions = 0;
int inserts = 0;
int deletes = 0;

extern "C" {
  typedef struct{
    int64_t value;
    int64_t next;
  } t_elem;
}

typedef struct {
  myMPI_Win twin; // table window
  myMPI_Win lwin; // last entry window
  myMPI_Win nwin; // nextfree window
  t_elem *table; // the table + heap
  int64_t nextfree; // the next free item in heap
  int64_t size; // size of the heap
  int64_t *last; // direct link to last element -- avoid traversing list!
  int p, r;
  int64_t tsize; // size of table
} t_hash;

int64_t hashfunc(int64_t val) {
  return val & (TSIZE-1);
}

void printHTAll(t_hash* hash, int n) {
 
  int val = -1;
  int nxt = -1;
  int owner = 0;
  int offset = 0;
  fprintf(stderr,"[%i] @@@@@@ Table: \n ",hash->r);
  for(int i = 0; i < TSIZE; i++) {
    if(i % hash->p == 0 && i > 0) {
      owner++;
      offset = 0;
      fprintf(stderr,"--\n");
    }

    myMPI_Get(&val,1,MPI_INT,owner,offset,1,MPI_INT,hash->twin);
    myMPI_Win_flush(owner, hash->twin);
    myMPI_Get(&nxt,1,MPI_INT,owner,offset+1,1,MPI_INT,hash->twin);
    myMPI_Win_flush(owner, hash->twin);
    fprintf(stderr,">> [%i] %i : (value: %i, next: %i)\n",owner,i,val,nxt);
    offset += 2;
  }

  owner = 0;
  offset = 0;

  fprintf(stderr,"\n[%i] @@@ Heap: \n",hash->r);
  for(int i = 0; i < n; i++) {

    if(i % hash->p == 0 && i > 0) {
      owner++;
      offset = 0;
      fprintf(stderr,"--\n");
    }

    myMPI_Get(&val,1,MPI_INT,owner,hash->tsize*2 + offset,1,MPI_INT,hash->twin);
    myMPI_Win_flush(owner, hash->twin);
    myMPI_Get(&nxt,1,MPI_INT,owner,hash->tsize*2 + offset+1,1,MPI_INT,hash->twin);
    myMPI_Win_flush(owner, hash->twin);
    fprintf(stderr,">> [%i] %i : (value: %i, next: %i)\n",owner,i+TSIZE,val,nxt);
    offset += 2;

  } 
}

void printHTLocal(t_hash* hash) {
  printf("[%i] @@@@@@ Table: \n ",hash->r);
  for(int i = 0; i < hash->tsize; i++) {    
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i].value, (hash->table)[i].next);
  }

  printf("\n[%i] @@@ Heap: \n",hash->r);
  for(int i = hash->tsize; i < hash->tsize+hash->size; i++) {
    printf("[%i] %i : (value: %i, next: %i)\n",hash->r,i,(hash->table)[i].value, (hash->table)[i].next); 
  }
}

void insert(t_hash* hash, int64_t elem) {
  int64_t pos = hashfunc(elem);
  int64_t owner = pos/hash->tsize;
  //printf("[%i] insert %i (pos: %i) - owner %i\n", hash->r, elem, pos, owner);  

  /*  if(owner == hash->r) { //local update
    // convert to local position
    pos = pos-hash->tsize*hash->r;
    
    if(hash->table[pos].value == -1) {
      hash->table[pos].value = elem;
    } 
    else {
      int newelem=hash->nextfree++;
      if(newelem >= hash->tsize+hash->size) {
	return;
      }
      if(hash->table[pos].next == -1) {
        // link new elem from table
        hash->table[pos].next = newelem;
      } 
      else {
        int newpos=hash->last[pos];
        hash->table[newpos].next = newelem;
      }
      hash->last[pos]=newelem;
      hash->table[newelem].value = elem;
    }
    } else*/  { // remote update
    // CAS if table position is available
    int64_t compare = -1;
    int64_t result = -2;
    int64_t offset /* in ints */ = (pos-owner*hash->tsize)*2;    

    myMPI_Compare_and_swap(&elem, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->twin);
    myMPI_Win_flush_local(owner, hash->twin);
    
    if(result != compare) { // we lost the CAS, no direct add
      ++insert_collisions;

      int64_t offset = (pos-owner*hash->tsize);
      int64_t origin = 1; // add 1
      int64_t result = -1; // address of newly added record

      // grab remote location
      myMPI_Fetch_and_op(/* origin */ &origin, /* result */ &result, MPI_INT64_T, owner, /* disp */ 0, myMPI_SUM, hash->nwin);
      myMPI_Win_flush_local(owner, hash->nwin);
      
      if(result >= hash->tsize+hash->size) {
        fprintf(stderr,"Failed to insert due to the OVERFLOW\n");
	return;
      }

      // put element in grabbed location
      myMPI_Put(&elem, 1, MPI_INT64_T, owner, 2*(result), 1, MPI_INT64_T, hash->twin);
      //myMPI_Win_flush(owner, hash->twin);
      // try to put directly as table link
      //elem = result;
      int64_t newOffset = result;
      
      int64_t oldOffset = -111;    /////mine
      offset = (pos-owner*hash->tsize);

      // grab lastptr element and plug in ptr to new element
      myMPI_Fetch_and_op(/* origin */ &newOffset, /* result */ &oldOffset, MPI_INT64_T, owner, /* disp */ offset, myMPI_REPLACE, hash->lwin);
      myMPI_Win_flush_local(owner, hash->lwin);

      compare = -1;
      result = -2;
      offset /* in ints */ = (pos-owner*hash->tsize)*2+1;

      // try to update table pointer directly (if it's the first update)
      myMPI_Compare_and_swap(&newOffset, &compare, &result, MPI_INT64_T, owner, /* target disp */ offset, hash->twin);
      myMPI_Win_flush_local(owner, hash->twin);

      if(result != compare) { // we lost the CAS for the direct pointer -- no direct access to table!
        // thus, insert 
        //origin = newoffset;
        //result = -2;
        //offset = (pos-owner*hash->tsize);
	//int64_t newOffset = -111;
        
        myMPI_Put(&newOffset, 1, MPI_INT64_T, owner, 2*(oldOffset)+1, 1, MPI_INT64_T, hash->twin);
        myMPI_Win_flush(owner, hash->twin);
      }
    }
  }
}

int64_t deleteElem(t_hash* hash, int64_t elem) {
  int64_t pos = hashfunc(elem);
  int64_t owner = pos/hash->tsize;
  int64_t posLocal = pos-owner*hash->tsize;
 
  /*if(owner == hash->r) { //local delete
    if(hash->table[posLocal].next == -1) {
      hash->table[posLocal].value = -1;
    }
    else {
      int offset = hash->last[posLocal];
      int prevOffset = posLocal;

      int nextOffset;
      int nextValue;

      int cnt = 2;
      while(cnt > 0) {
	nextOffset = hash->table[prevOffset].next;
	nextValue = hash->table[nextOffset].value;

	if(nextOffset == -1 || nextValue == SENTINEL) {
	  return -1;
	}

	if(nextOffset == offset) {
	  hash->table[prevOffset].next = -1;
	  hash->last[posLocal] = prevOffset;
	  hash->table[offset].value = -1;

	  return nextValue;
	}

	prevOffset = hash->table[prevOffset].next;
	cnt--;
      }
    }
  }
  else*/  { //remote delete
    int64_t compare = -1;
    int64_t result = -2;
    int64_t offsetValue = (pos-owner*hash->tsize);
    int64_t next = -11;
    int64_t newValue = -1;

    //get the next pointer of the element at the position offsetValue.
    myMPI_Get(&next, 1, MPI_INT64_T, owner, 2*offsetValue+1, 1, MPI_INT64_T, hash->twin);
    myMPI_Win_flush(owner, hash->twin);

    //if the next element is equal to -1 then simply erase this element from the table
    //SHIT!!! It's not safe. How to use CAS on the value and the next simultaneously? locks...?
    if(next == -1) {
      myMPI_Put(&newValue, 1, MPI_INT64_T, owner, 2*offsetValue, 1, MPI_INT64_T, hash->twin);
      //myMPI_Win_flush(owner, hash->twin);
    }
    else {
      //otherwise let's chase the element...
      ++delete_collisions;
      int64_t offset = -5;
      int64_t prevOffset = offsetValue;
      int64_t nextOffset = -6;
      //int64_t nextValue = -7;

      //obtain the pointer to the last element
      myMPI_Get(&offset, 1, MPI_INT64_T, owner, pos-owner*hash->tsize, 1, MPI_INT64_T, hash->lwin);
      myMPI_Win_flush(owner, hash->lwin);
      
      //traverse the list until we find the element before the last one
      while(1) {
	//get the next offset
	myMPI_Get(&nextOffset, 1, MPI_INT64_T, owner, 2*prevOffset+1, 1, MPI_INT64_T, hash->twin);
	myMPI_Win_flush(owner, hash->twin);

	//myMPI_Get(&nextValue, 1, MPI_INT64_T, owner, 2*nextOffset, 1, MPI_INT64_T, hash->twin);
	//myMPI_Win_flush(owner, hash->twin);

	//exit if we went too far...
	if(nextOffset == -1 || nextOffset >= hash->tsize+hash->size || nextOffset <= prevOffset) {
	  fprintf(stderr,"Failed to delete\n");
	  return -1;
	}

	//if the current value of nextOffset matches the value of the last pointer, then we have it
	if(nextOffset == offset) {

	  //update the last pointer (from the 'last' array)
          myMPI_Put(&prevOffset, 1, MPI_INT64_T, owner, offsetValue, 1, MPI_INT64_T, hash->lwin);
          myMPI_Win_flush_local(owner, hash->lwin);

	  //set 'next' pointer of the previous elem to -1 (newValue) 
	  myMPI_Put(&newValue, 1, MPI_INT64_T, owner, 2*prevOffset+1, 1, MPI_INT64_T, hash->twin);
	  //myMPI_Win_flush(owner, hash->twin);

	  //set the value to be deleted to -1
	  myMPI_Put(&newValue, 1, MPI_INT64_T, owner, 2*offset, 1, MPI_INT64_T, hash->twin);
	  myMPI_Win_flush_local(owner, hash->twin);
	  
	  return 1;//nextValue;
	}

	//get the next pointer
	myMPI_Get(&prevOffset, 1, MPI_INT64_T, owner, 2*prevOffset+1, 1, MPI_INT64_T, hash->twin);
	myMPI_Win_flush(owner, hash->twin);
      }
    }
  }
}

int count(t_hash *hash) {

  int cnt=0;
  int heapcnt=0;
  for(int i=0; i<hash->tsize; i++) {
    int pos = i;
    if(hash->table[pos].value != -1) {
      cnt++;
      if(hash->table[pos].next != -1) {
        heapcnt++;
        pos=hash->table[pos].next;
        while(hash->table[pos].next != -1) {
          pos = hash->table[pos].next;
          heapcnt++;
        }
      }
    }
  }
  //  printf("cnt: %i heapcnt: %i, total: %i\n", cnt, heapcnt, cnt+heapcnt);
  return cnt+heapcnt;
}

int countSimple(t_hash* hash) {
  int cnt = 0;
  for(int i = 0; i < hash->tsize+hash->size;i++) {
    if(hash->table[i].value != -1) {
      cnt++;
    }
  }
  return cnt;
}

int main(int argc, char **argv) {
  myMPI_Init(&argc, &argv);
  
  t_hash hash;
  int LV = atoi(argv[1]);
  
  MPI_Comm_size(MPI_COMM_WORLD, &hash.p);
  MPI_Comm_rank(MPI_COMM_WORLD, &hash.r);
  
  TSIZE = LV*hash.p;
  srand(hash.r*1000000);
  //assert(RAND_MAX > TSIZE);
  
  hash.tsize = LV;
  hash.size = (MAX_AVG_COLLISIONS)*LV;

  myMPI_Win_allocate((hash.tsize+hash.size)*sizeof(t_elem), sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.table, &hash.twin);
  myMPI_Win_lock_all(0, hash.twin);

  myMPI_Win_allocate(hash.tsize*sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, MPI_COMM_WORLD, &hash.last, &hash.lwin);
  myMPI_Win_lock_all(0, hash.lwin);

  hash.nextfree = hash.tsize; // next free element in heap
  myMPI_Win_create(&hash.nextfree, sizeof(int64_t), 1, MPI_INFO_NULL, MPI_COMM_WORLD, &hash.nwin);
  myMPI_Win_lock_all(0, hash.nwin);

  //MTRand mt(999-hash.r);

  // initialize table
  for(int i=0; i<hash.tsize+hash.size; ++i) {
    if(i<hash.tsize) hash.last[i]=-1;
    hash.table[i].next=-1;
    hash.table[i].value=-1;
  }

  double t_start = 0.0;
  double t_end = 0.0;
  
  int nr_of_ins = HTABLE_FILL_RATIO*LV;
  if(hash.r == 0) printf("nr of ins: %i\n",nr_of_ins);

  int warmups = 0.1*nr_of_ins;
  double ins_time = 0.0;
  if(hash.r == 0) printf("warmups: %i\n",warmups);

  int64_t* vals = (int64_t*)calloc(nr_of_ins,sizeof(int64_t));

  srand(hash.r*1000000);

  MPI_Barrier(MPI_COMM_WORLD);

  for(int m = 0; m < nr_of_ins; m++) {
    if(m == warmups) {
      MPI_Barrier(MPI_COMM_WORLD);
      t_start = MPI_Wtime();
    }
    int64_t val = bigRandVal();
    insert(&hash, val);
    inserts++;
    vals[m] = val;
  }

  MPI_Barrier(MPI_COMM_WORLD);
  t_end = MPI_Wtime();
  ins_time = t_end - t_start;

  if(hash.r == 0) {
     printf("%i\t%f\n",hash.p,ins_time);
   }

  printf("[%i] collisions: %i\n",hash.r,insert_collisions );

  int size = count(&hash);
  assert(size == countSimple(&hash));
  //printf("[%i] cnt: %i, simpleCnt:%i\n",hash.r,size,countSimple(&hash));

  MPI_Barrier(MPI_COMM_WORLD);

  int sum = -1;
  MPI_Reduce(&size, &sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

   if(hash.r == 0) {
    assert(nr_of_ins * hash.p == sum);
    printf("The sum: %i\n", sum);
  }

  int64_t* elems = (int64_t*)calloc(size,sizeof(int64_t));
  int b = 0;
  for(int i = 0; i < hash.size+hash.tsize;i++) {
    if(hash.table[i].value != -1) {
      elems[b] = hash.table[i].value;
      b++;
    }
  }

  int64_t* generated[hash.p];
   int64_t* received[hash.p];
  int lengths[hash.p];

  if(hash.r != 0) {
    //MPI_Send(elems,size,MPI_INT64_T,0,0,MPI_COMM_WORLD);
    MPI_Send(vals,nr_of_ins,MPI_INT64_T,0,0,MPI_COMM_WORLD);
  }
  else {

    generated[hash.r] = vals;
    
    int procs_done = 1;
    int flag = false;
    MPI_Status s;

    while(procs_done < hash.p) {
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &s);

      if(flag == true) {
	flag = false;
	int sender = s.MPI_SOURCE;
	int tag = s.MPI_TAG;
	int len = -1;
	MPI_Get_count(&s, MPI_INT64_T, &len);
	int64_t* data = (int64_t*)calloc(len,sizeof(int64_t));
	MPI_Recv(data, len, MPI_INT64_T, sender, tag, MPI_COMM_WORLD, &s);
	procs_done++;
	generated[sender] = data;
	assert(len == nr_of_ins);
      }
    }
  }
    
  MPI_Barrier(MPI_COMM_WORLD);

  if(hash.r != 0) {
    MPI_Send(elems,size,MPI_INT64_T,0,0,MPI_COMM_WORLD);
  }
  else {
    received[hash.r] = elems;
    lengths[hash.r] = size;

    int procs_done = 1;
    int flag = false;
    MPI_Status s;

    while(procs_done < hash.p) {
      MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &flag, &s);

      if(flag == true) {
        flag = false;
        int sender = s.MPI_SOURCE;
        int tag = s.MPI_TAG;
        int len = -1;
        MPI_Get_count(&s, MPI_INT64_T, &len);
        int64_t* data = (int64_t*)calloc(len,sizeof(int64_t));
        MPI_Recv(data, len, MPI_INT64_T, sender, tag, MPI_COMM_WORLD, &s);
	procs_done++;
        received[sender] = data;
        lengths[sender] = len;
      }
    }
  }

  MPI_Barrier(MPI_COMM_WORLD);

  if(hash.r == 0) {
    int sum_of_generated = hash.p*nr_of_ins;
    int sum_of_received = 0;

    for(int i = 0; i < hash.p; i++) {
      sum_of_received += lengths[i];
    }

    assert(sum_of_generated == sum_of_received);

    int matching = 0;
    for(int i = 0; i < hash.p; i++) {
      for(int k = 0; k < nr_of_ins; k++) {
	
	int64_t investigated = generated[i][k];
	//printf("\n\n>>>> looking for.. %llu\n\n", generated[i][k]);
	//int found = -1;
	
	for(int m = 0; m < hash.p; m++) {
	  for(int n = 0; n < lengths[m]; n++) {
	    //printf("%llu --- ",received[m][n]);
	    if(received[m][n] == investigated) {
	      received[m][n] = -1;
	      matching++;
	      //printf("found %i\n",matching);
	      goto next_check;
	    }
	  }
	}
	assert(1 == 0);
      next_check:
      }
    }
    assert(matching == sum_of_received);
  }

    
  /*    printf(">>>>>>>>>>>>>>>>>>>>> vals: \n");
    for(int i = 0; i < hash.p; i++) {
      for(int k = 0; k < nr_of_ins; k++) {
	printf("%llu --- ",generated[i][k]);
      }
      printf("\n===\n");
    }
    printf(">>>>>>>>>>>>>>>>>>>>> elems: \n");
    for(int i = 0; i < hash.p; i++) {
      for(int k = 0; k < lengths[i]; k++) {
        printf("%llu --- ",received[i][k]);
      }
      printf("\n==\n");
      }*/

  myMPI_Finalize();
}
