#ifndef BENCHMARK_FOMPI_EX
#define BENCHMARK_FOMPI_EX

#include <stdio.h>    // Standard stuff
#include <stdlib.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include <string>

#include <inttypes.h>
#include <stdint.h>

#include "fompi.h"
#include "Locks.h"

#define SLEEP_TIME 4
#define MAX_ITERATIONS 100000
#define ITERATIONS_PER_WARMUP 10000
#define USECONDS_FOR_THROUGHPUT 1000000.0

#define MCS_LATENCY 1
#define MCS_SINGLE_MEMORY_READ 1
#define MCS_THROUGHPUT_NO_CS 1
#define MCS_THROUGHPUT_SLEEP_IN_CS 1
#define MCS_THROUGHPUT_SLEEP_AFTER_CS 1

#define HMCS_LATENCY 1
#define HMCS_SINGLE_MEMORY_READ 1
#define HMCS_THROUGHPUT_NO_CS 1
#define HMCS_THROUGHPUT_SLEEP_IN_CS 1
#define HMCS_THROUGHPUT_SLEEP_AFTER_CS 1

#define FOMPI_LATENCY 1
#define FOMPI_SINGLE_MEMORY_READ 1
#define FOMPI_THROUGHPUT_NO_CS 1
#define FOMPI_THROUGHPUT_SLEEP_IN_CS 1
#define FOMPI_THROUGHPUT_SLEEP_AFTER_CS 1

#define RW_LATENCY 1
#define RW_SINGLE_MEMORY_READ 1
#define RW_THROUGHPUT_NO_CS 1
#define RW_THROUGHPUT_SLEEP_IN_CS 1
#define RW_THROUGHPUT_SLEEP_AFTER_CS 1

#define THRESHOLD_RW_LATENCY 1
#define THRESHOLD_RW_SINGLE_MEMORY_READ 1
#define THRESHOLD_RW_THROUGHPUT_NO_CS 1
#define THRESHOLD_RW_THROUGHPUT_SLEEP_IN_CS 1
#define THRESHOLD_RW_THROUGHPUT_SLEEP_AFTER_CS 1

#define RW_FOMPI_LATENCY 1
#define RW_FOMPI_SINGLE_MEMORY_READ 1
#define RW_FOMPI_THROUGHPUT_NO_CS 1
#define RW_FOMPI_THROUGHPUT_SLEEP_IN_CS 1
#define RW_FOMPI_THROUGHPUT_SLEEP_AFTER_CS 1

extern int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

extern MPI_File file_latency, file_throughput_sleep_in_CS, file_throughput_sleep_after_CS, file_throughput_no_CS, file_single_memory_read;

void mcs_latency(MPI_Comm comm, MPI_File filehandle, MCSLock mcs_lock);
void mcs_single_memory_read(MPI_Comm comm, MPI_File filehandle, MCSLock mcs_lock);
void mcs_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle, MCSLock mcs_lock);
void mcs_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle, MCSLock mcs_lock);
void mcs_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle, MCSLock mcs_lock);

void hmcs_latency(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock);
void hmcs_single_memory_read(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock);
void hmcs_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock);
void hmcs_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock);
void hmcs_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock);

void fompi_latency(MPI_Comm comm, MPI_File filehandle);
void fompi_single_memory_read(MPI_Comm comm, MPI_File filehandle);
void fompi_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle);
void fompi_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle);
void fompi_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle);

void rw_latency(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void rw_single_memory_read(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void rw_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void rw_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void rw_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);

void rw_fompi_latency(MPI_Comm comm, MPI_File filehandle);
void rw_fompi_single_memory_read(MPI_Comm comm, MPI_File filehandle);
void rw_fompi_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle);
void rw_fompi_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle);
void rw_fompi_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle);

void threshold_rw_latency(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void threshold_rw_single_memory_read(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void threshold_rw_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void threshold_rw_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);
void threshold_rw_throughput_sleep_after_CS(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock);

inline void open_files(MPI_Comm comm, string lock_name, bool latency, bool single_memory_read, bool throughput_empty_CS, bool throughput_sleep_after_CS, bool throughput_sleep_in_CS) {
	
	int rank, ier, comm_size;
	char filename[80];
	char timestr[25];
	
	time_t mytime;
	struct tm* tm_info;
	
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	
	/* Initialization of output file */
	mytime = time(NULL);
	tm_info = localtime(&mytime);
	strftime(timestr, 25, "%Y-%m-%d_%H-%M", tm_info);
	if(latency && rank == 0){
		snprintf(&filename[0], 80, "%s-latency-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_latency);
	
		if (ier != MPI_SUCCESS) {
			printf(
					"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(single_memory_read && rank == 0){
		snprintf(&filename[0], 80, "%s-single_memory_read-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_single_memory_read);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(throughput_empty_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-emptyCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_no_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	if(throughput_sleep_after_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-sleepAfterCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_after_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(throughput_sleep_in_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-sleepInCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_in_CS);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
}

#endif
