#include "benchmark.h"

unsigned long long g_timerfreq;
double g_hrttimer_startvalue;

MPI_File file_latency, file_throughput_sleep_in_CS, file_throughput_sleep_after_CS, file_throughput_no_CS, file_single_memory_read;

int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

using namespace std;

/*
 * benchmark of HMCS Lock
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
int main(int argc, char *argv[]) {
	int rank, comm_size;
	MPI_Comm comm;
	
	HMCSLock hmcs_lock;
	
	marker_dist = atoi(argv[1]);
	hmcs_threshold = atoi(argv[2]);
	tree_threshold = atoi(argv[3]);
	root_threshold = atoi(argv[4]);
	reader_threshold = atoi(argv[5]);
	writer_rate = atoi(argv[6]);
	
	srand(time(NULL) + rank * 100000);
	HRT_INIT(1, g_timerfreq);

	foMPI_Init(&argc, &argv);
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	
	if(rank == 0){
		open_files(comm, "hmcs", HMCS_LATENCY, HMCS_SINGLE_MEMORY_READ, HMCS_THROUGHPUT_NO_CS, HMCS_THROUGHPUT_SLEEP_AFTER_CS, HMCS_THROUGHPUT_SLEEP_IN_CS);
	}
	
	hmcs_lock.initialize();
	
	if(HMCS_LATENCY){
		hmcs_latency(comm, file_latency, hmcs_lock);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(HMCS_SINGLE_MEMORY_READ){
		hmcs_single_memory_read(comm, file_single_memory_read, hmcs_lock);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(HMCS_THROUGHPUT_NO_CS){
		hmcs_throughput_empty_CS(comm, file_throughput_no_CS, hmcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(HMCS_THROUGHPUT_SLEEP_AFTER_CS){
		hmcs_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS, hmcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(HMCS_THROUGHPUT_SLEEP_IN_CS){
		hmcs_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS, hmcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	
	hmcs_lock.finalize();
	foMPI_Finalize();
	
	return 0;
}
