#include "benchmark.h"

unsigned long long g_timerfreq;
double g_hrttimer_startvalue;

MPI_File file_latency, file_throughput_sleep_in_CS, file_throughput_sleep_after_CS, file_throughput_no_CS, file_single_memory_read;

int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

using namespace std;

/*
 * benchmark of RW Lock for writer_rate 25%
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
int main(int argc, char *argv[]) {
	int rank, comm_size;
	MPI_Comm comm;
	
	RWLock rw_lock;
	
	marker_dist = atoi(argv[1]);
	hmcs_threshold = atoi(argv[2]);
	tree_threshold = atoi(argv[3]);
	root_threshold = atoi(argv[4]);
	reader_threshold = atoi(argv[5]);
	writer_rate = atoi(argv[6]);
	
	srand(time(NULL) + rank * 100000);
	HRT_INIT(1, g_timerfreq);
	
	string lock_name = "rw-25-" + std::to_string(tree_threshold) + "-" + std::to_string(root_threshold);

	foMPI_Init(&argc, &argv);
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	
	if(rank == 0){
		open_files(comm, lock_name, THRESHOLD_RW_LATENCY, THRESHOLD_RW_SINGLE_MEMORY_READ, THRESHOLD_RW_THROUGHPUT_NO_CS, THRESHOLD_RW_THROUGHPUT_SLEEP_AFTER_CS, THRESHOLD_RW_THROUGHPUT_SLEEP_IN_CS);
	}
	
	rw_lock.initialize();
	
	if(THRESHOLD_RW_LATENCY){
		threshold_rw_latency(comm, file_latency, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(THRESHOLD_RW_SINGLE_MEMORY_READ){
		threshold_rw_single_memory_read(comm, file_single_memory_read, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_NO_CS){
		threshold_rw_throughput_empty_CS(comm, file_throughput_no_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_SLEEP_AFTER_CS){
		threshold_rw_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_SLEEP_IN_CS){
		threshold_rw_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	
	rw_lock.finalize();
	foMPI_Finalize();
	
	return 0;
}
