#include "benchmark.h"

using namespace std;

/*
 * Measuring throughput of the foMPI_LOCK_EXCLUSIVE with a given contention 
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void fompi_contention(MPI_Comm comm, MPI_File filehandle) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long ticks_diff;
	
	char line[256];
	double *throughputs;
	double *tot_attempts;
	int iterations = ITERATIONS_FOR_THROUGHPUT;
	
	foMPI_Win win_count, win_results;
	int comm_size;
	int rank;
	int one = 1;
	
	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(iterations * sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &tot_attempts, &win_count);
	foMPI_Win_allocate(iterations * sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &throughputs, &win_results);

	printf("%i before warm-up\n", rank);
	/* Warm-up */
	int i;
	double dummy;
	for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
		foMPI_Win_lock(&dummy, foMPI_LOCK_EXCLUSIVE, 0, 0, win_count);
		foMPI_Win_unlock(0, win_count);
	}
	
	printf("%i before test\n", rank);
	/* Start test */
	for(i = 0; i < iterations; i++){
		double lock_count = 0;
		double time_diff = 0;
		double attempts = 0;
		MPI_Barrier(comm);
		HRT_GET_TIMESTAMP(t[0]);
		/* Execute while loop until 1 second is reached */
		while(time_diff < USECONDS_FOR_THROUGHPUT){
			foMPI_Win_lock(&attempts, foMPI_LOCK_EXCLUSIVE, 0, 0, win_count);
			foMPI_Win_unlock(0, win_count);
			
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &ticks_diff);
			time_diff = HRT_GET_USEC(ticks_diff);
			lock_count++;
		
			/* Sleep for 1 to 4 us */
			HRT_TIMESTAMP_T time_stamp[2];
			unsigned long long diff;
			double time_wait = 0;
			int sleep_time = SLEEP_TIME;
			HRT_GET_TIMESTAMP(time_stamp[0]);
			while(time_wait < sleep_time)
			{
				HRT_GET_TIMESTAMP(time_stamp[1]);
				HRT_GET_ELAPSED_TICKS(time_stamp[0], time_stamp[1], &diff);
				time_wait = HRT_GET_USEC(diff);
			}		
		}
		throughputs[i] = lock_count / (time_diff / 1000000.0);
		tot_attempts[i] = attempts / (time_diff / 1000000.0);
	}

	MPI_Barrier(comm);
	if(rank == 0){
		int j;
		/* Get the count of throughput from each process */
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			double buf_throughput[ITERATIONS_FOR_THROUGHPUT];
			double buf_attempts[ITERATIONS_FOR_THROUGHPUT];
			foMPI_Get(buf_throughput, iterations, MPI_DOUBLE, j, 0, iterations, MPI_DOUBLE, win_results);
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_count);
			foMPI_Get(buf_attempts, iterations, MPI_DOUBLE, j, 0, iterations, MPI_DOUBLE, win_count);
			foMPI_Win_fence(0, win_count);
			int k;
			for(k = 0; k < iterations; k++){
				throughputs[k] += buf_throughput[k];
				tot_attempts[k] += buf_attempts[k];
			}
		}
		/* Write result to a file */
		int l;
		double total_count = 0, total_attempts = 0;
		for (l = 0; l < iterations; l++) {
			total_count += throughputs[l];
			total_attempts += tot_attempts[l];
		}
		snprintf(&line[0], 256, "%20.4f\t%20.4f", total_count / iterations, total_attempts / iterations);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_count);
			foMPI_Win_fence(0, win_count);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win_count);
	foMPI_Win_free(&win_results);
}
