#include "benchmark.h"

using namespace std;

/*
 * Measuring throughput of the HMCS-lock with global count in CS and sleeping in CS from 1 to 4 us 
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void hmcs_throughput_sleep_in_CS(MPI_Comm comm, MPI_File filehandle, HMCSLock hmcs_lock) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long ticks_diff;
	double time_diff;
	
	/* Timer for sleeping */
	HRT_TIMESTAMP_T time_stamp[2];
	unsigned long long diff;
	double time_wait;
	int sleep_time;
	
	char line[256];
	double *throughput;
	
	foMPI_Win win_count, win_results;
	int comm_size;
	int rank;
	int64_t one = 1, *count;
	
	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &count, &win_count);
	foMPI_Win_allocate(sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &throughput, &win_results);
	*count = 0;
	
	/* Warm-up */
	int i;
	for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
		hmcs_lock.acquire();
		foMPI_Accumulate(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, foMPI_SUM, win_count);
		foMPI_Win_flush(0, win_count);
		hmcs_lock.release();
	}
			
	foMPI_Win_fence(0, win_count);
	HRT_GET_TIMESTAMP(t[0]);
	/* Start test */
	for(i = 0; i < MAX_ITERATIONS; i++){
		hmcs_lock.acquire();
		foMPI_Accumulate(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, foMPI_SUM, win_count);
		foMPI_Win_flush(0, win_count);
				
		/* Sleep for 0 to 4 us */
		HRT_GET_TIMESTAMP(time_stamp[0]);
		sleep_time = rand() % SLEEP_TIME;
		while(time_wait < sleep_time)
		{
			HRT_GET_TIMESTAMP(time_stamp[1]);
			HRT_GET_ELAPSED_TICKS(time_stamp[0], time_stamp[1], &diff);
			time_wait = HRT_GET_USEC(diff);
		}
		
		hmcs_lock.release();
	}
	
	HRT_GET_TIMESTAMP(t[1]);
	HRT_GET_ELAPSED_TICKS(t[0], t[1], &ticks_diff);
	time_diff = HRT_GET_USEC(ticks_diff);
	*throughput = MAX_ITERATIONS / (time_diff / USECONDS_FOR_THROUGHPUT);

	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		int j;
		/* Get the count of throughput from each process */
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			double buf;
			foMPI_Get(&buf, 1, MPI_DOUBLE, j, 0, 1, MPI_DOUBLE, win_results);
			foMPI_Win_fence(0, win_results);
			*throughput += buf;
		}
		/* Write result to a file */
		snprintf(&line[0], 256, "%20.4f", *throughput);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_results);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win_count);
	foMPI_Win_free(&win_results);
}

