#include <stdio.h>
#include <time.h>
#include <string>

#include "benchmark.h"

using namespace std;

unsigned long long g_timerfreq;
double g_hrttimer_startvalue;

/*
 * Main function for microbenchmarks
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
int main(int argc, char *argv[]) {

	int rank, comm_size;
	MPI_Comm comm;
	MCSLock mcs_lock;
	
	MPI_File file_latency, file_throughput_sleep_in_CS, file_throughput_sleep_after_CS, file_throughput_no_CS, file_single_memory_read;
		
	int ier;
	char filename[80];
	char timestr[25];
	string lock_name;
	time_t mytime;
	struct tm* tm_info;
	
	RWLock rw_lock;
	
	foMPI_Init(&argc, &argv);
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);

	srand(time(NULL) + rank * 100000);
	HRT_INIT(1, g_timerfreq);

	/* Initialization of output file */
	mytime = time(NULL);
	tm_info = localtime(&mytime);
	strftime(timestr, 25, "%Y-%m-%d_%H-%M", tm_info);
	if((MCS_LATENCY || FOMPI_LATENCY || RW_LATENCY || RW_FOMPI_LATENCY || THRESHOLD_RW_LATENCY) && rank == 0){
		if(MCS_LATENCY){
			lock_name = "mcs";
		} else if(FOMPI_LATENCY){
			lock_name = "fompi";
		} else if(RW_LATENCY){
			lock_name = "rw-m" + std::to_string(MARKER_DIST) + "-r" + std::to_string(READER_THRESHOLD) + "-r" + std::to_string(WRITER_RATE / 10);
		} else if(RW_FOMPI_LATENCY){
			lock_name = "rw-fompi-" + std::to_string(WRITER_RATE / 10);
		} else {
			lock_name = "rw-25-" + std::to_string(TREE_THRESHOLD) + "-" + std::to_string(WRITER_ROOT_THRESHOLD);
		}
		snprintf(&filename[0], 80, "%s-latency-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_latency);
	
		if (ier != MPI_SUCCESS) {
			printf(
					"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if((MCS_SINGLE_MEMORY_READ || FOMPI_SINGLE_MEMORY_READ || RW_SINGLE_MEMORY_READ || RW_FOMPI_SINGLE_MEMORY_READ || THRESHOLD_RW_SINGLE_MEMORY_READ) && rank == 0){
		if(MCS_SINGLE_MEMORY_READ){
			lock_name = "mcs";
		} else if(FOMPI_SINGLE_MEMORY_READ){
			lock_name = "fompi";
		} else if(RW_SINGLE_MEMORY_READ){
			lock_name = "rw-m" + std::to_string(MARKER_DIST) + "-r" + std::to_string(READER_THRESHOLD) + "-r" + std::to_string(WRITER_RATE / 10);
		} else if(RW_FOMPI_SINGLE_MEMORY_READ){
			lock_name = "rw-fompi-" + std::to_string(WRITER_RATE / 10);
		} else {
			lock_name = "rw-25-" + std::to_string(TREE_THRESHOLD) + "-" + std::to_string(WRITER_ROOT_THRESHOLD);
		}
		snprintf(&filename[0], 80, "%s-single_memory_read-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_single_memory_read);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if((MCS_THROUGHPUT_NO_CS || FOMPI_THROUGHPUT_NO_CS || RW_THROUGHPUT_NO_CS || RW_FOMPI_THROUGHPUT_NO_CS || THRESHOLD_RW_THROUGHPUT_NO_CS) && rank == 0){
		if(MCS_THROUGHPUT_NO_CS){
			lock_name = "mcs";
		} else if(FOMPI_THROUGHPUT_NO_CS){
			lock_name = "fompi";
		} else if(RW_THROUGHPUT_NO_CS){
			lock_name = "rw-m" + std::to_string(MARKER_DIST) + "-r" + std::to_string(READER_THRESHOLD) + "-r" + std::to_string(WRITER_RATE / 10);
		} else if(RW_FOMPI_THROUGHPUT_NO_CS){
			lock_name = "rw-fompi-" + std::to_string(WRITER_RATE / 10);
		} else {
			lock_name = "rw-25-" + std::to_string(TREE_THRESHOLD) + "-" + std::to_string(WRITER_ROOT_THRESHOLD);
		}
		snprintf(&filename[0], 80, "%s-throughput-emptyCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_no_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	if((MCS_THROUGHPUT_SLEEP_AFTER_CS || FOMPI_THROUGHPUT_SLEEP_AFTER_CS || RW_THROUGHPUT_SLEEP_AFTER_CS || RW_FOMPI_THROUGHPUT_SLEEP_AFTER_CS ||THRESHOLD_RW_THROUGHPUT_SLEEP_AFTER_CS) && rank == 0){
		if(MCS_THROUGHPUT_SLEEP_AFTER_CS){
			lock_name = "mcs";
		} else if(FOMPI_THROUGHPUT_SLEEP_AFTER_CS){
			lock_name = "fompi";
		} else if(RW_THROUGHPUT_SLEEP_AFTER_CS){
			lock_name = "rw-m" + std::to_string(MARKER_DIST) + "-r" + std::to_string(READER_THRESHOLD) + "-r" + std::to_string(WRITER_RATE / 10);
		} else if(RW_FOMPI_THROUGHPUT_SLEEP_AFTER_CS){
			lock_name = "rw-fompi-" + std::to_string(WRITER_RATE / 10);
		} else {
			lock_name = "rw-25-" + std::to_string(TREE_THRESHOLD) + "-" + std::to_string(WRITER_ROOT_THRESHOLD);
		}
		snprintf(&filename[0], 80, "%s-throughput-sleepAfterCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_after_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if((MCS_THROUGHPUT_SLEEP_IN_CS || FOMPI_THROUGHPUT_SLEEP_IN_CS || RW_THROUGHPUT_SLEEP_IN_CS || RW_FOMPI_THROUGHPUT_SLEEP_IN_CS || THRESHOLD_RW_THROUGHPUT_SLEEP_IN_CS) && rank == 0){
		if(MCS_THROUGHPUT_SLEEP_IN_CS){
			lock_name = "mcs";
		} else if(FOMPI_THROUGHPUT_SLEEP_IN_CS){
			lock_name = "fompi";
		} else if(RW_THROUGHPUT_SLEEP_IN_CS){
			lock_name = "rw-m" + std::to_string(MARKER_DIST) + "-r" + std::to_string(READER_THRESHOLD) + "-r" + std::to_string(WRITER_RATE / 10);
		} else if(RW_FOMPI_THROUGHPUT_SLEEP_IN_CS){
			lock_name = "rw-fompi-" + std::to_string(WRITER_RATE / 10);
		} else {
			lock_name = "rw-25-" + std::to_string(TREE_THRESHOLD) + "-" + std::to_string(WRITER_ROOT_THRESHOLD);
		}
		snprintf(&filename[0], 80, "%s-throughput-sleepInCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_in_CS);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	/* Execution of benchmarks */
	mcs_lock.initialize();
	if(MCS_LATENCY){
		mcs_latency(comm, file_latency, mcs_lock);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}	
	if(MCS_SINGLE_MEMORY_READ){
		mcs_single_memory_read(comm, file_single_memory_read, mcs_lock);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(MCS_THROUGHPUT_NO_CS){
		mcs_throughput_empty_CS(comm, file_throughput_no_CS, mcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}	
	if(MCS_THROUGHPUT_SLEEP_AFTER_CS){
		mcs_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS, mcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}	
	if(MCS_THROUGHPUT_SLEEP_IN_CS){
		mcs_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS, mcs_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	mcs_lock.finalize();
	
	if(FOMPI_LATENCY){
		fompi_latency(comm, file_latency);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(FOMPI_SINGLE_MEMORY_READ){
		fompi_single_memory_read(comm, file_single_memory_read);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(FOMPI_THROUGHPUT_NO_CS){
		fompi_throughput_empty_CS(comm, file_throughput_no_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(FOMPI_THROUGHPUT_SLEEP_AFTER_CS){
		fompi_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(FOMPI_THROUGHPUT_SLEEP_IN_CS){
		fompi_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
		
	rw_lock.initialize();
	if(RW_LATENCY){
		rw_latency(comm, file_latency, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(RW_SINGLE_MEMORY_READ){
		rw_single_memory_read(comm, file_single_memory_read, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(RW_THROUGHPUT_NO_CS){
		rw_throughput_empty_CS(comm, file_throughput_no_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(RW_THROUGHPUT_SLEEP_AFTER_CS){
		rw_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(RW_THROUGHPUT_SLEEP_IN_CS){
		rw_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	
	if(THRESHOLD_RW_LATENCY){
		threshold_rw_latency(comm, file_latency, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(THRESHOLD_RW_SINGLE_MEMORY_READ){
		threshold_rw_single_memory_read(comm, file_single_memory_read, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_NO_CS){
		threshold_rw_throughput_empty_CS(comm, file_throughput_no_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_SLEEP_AFTER_CS){
		threshold_rw_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(THRESHOLD_RW_THROUGHPUT_SLEEP_IN_CS){
		threshold_rw_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS, rw_lock);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	
	rw_lock.finalize();
	
	if(RW_FOMPI_LATENCY){
		rw_fompi_latency(comm, file_latency);
		if(rank == 0){
			MPI_File_close(&file_latency);
		}
	}
	if(RW_FOMPI_SINGLE_MEMORY_READ){
		rw_fompi_single_memory_read(comm, file_single_memory_read);
		if(rank == 0){
			MPI_File_close(&file_single_memory_read);
		}
	}
	if(RW_FOMPI_THROUGHPUT_NO_CS){
		rw_fompi_throughput_empty_CS(comm, file_throughput_no_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_no_CS);
		}
	}
	if(RW_FOMPI_THROUGHPUT_SLEEP_AFTER_CS){
		rw_fompi_throughput_sleep_after_CS(comm, file_throughput_sleep_after_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_after_CS);
		}
	}
	if(RW_FOMPI_THROUGHPUT_SLEEP_IN_CS){
		rw_fompi_throughput_sleep_in_CS(comm, file_throughput_sleep_in_CS);
		if(rank == 0){
			MPI_File_close(&file_throughput_sleep_in_CS);
		}
	}
	
	//foMPI_Finalize();
	return 0;
}
