#include "benchmark.h"

void open_files(MPI_Comm comm, string lock_name, bool latency, bool single_memory_read, bool throughput_empty_CS, bool throughput_sleep_after_CS, bool throughput_sleep_in_CS) {
	
	int rank, ier, comm_size;
	char filename[80];
	char timestr[25];
	
	time_t mytime;
	struct tm* tm_info;
	
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	
	srand(time(NULL) + rank * 100000);
	HRT_INIT(1, g_timerfreq);
	
	/* Initialization of output file */
	mytime = time(NULL);
	tm_info = localtime(&mytime);
	strftime(timestr, 25, "%Y-%m-%d_%H-%M", tm_info);
	if(latency && rank == 0){
		snprintf(&filename[0], 80, "%s-latency-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_latency);
	
		if (ier != MPI_SUCCESS) {
			printf(
					"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(single_memory_read && rank == 0){
		snprintf(&filename[0], 80, "%s-single_memory_read-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_single_memory_read);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(throughput_empty_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-emptyCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_no_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	if(throughput_sleep_after_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-sleepAfterCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_after_CS);
	
		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
	
	if(throughput_sleep_in_CS && rank == 0){
		snprintf(&filename[0], 80, "%s-throughput-sleepInCS-%ip.out", lock_name.c_str(), comm_size);
		ier = MPI_File_open( MPI_COMM_SELF, filename,
			MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &file_throughput_sleep_in_CS);

		if (ier != MPI_SUCCESS) {
			printf(
				"Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n",
				filename);
			MPI_Abort(comm, 1);
		}
	}
}