#include "benchmark.h"
#include <random>

using namespace std;

/*
 * Measuring throughput of the fompi's RW lock with one single memory operation in CS
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void rw_fompi_single_memory_read(MPI_Comm comm, MPI_File filehandle) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long ticks_diff;
	double time_diff;
	
	char line[256];
	double *throughput;
	
	foMPI_Win win_count, win_results;
	int comm_size;
	int rank;
	int64_t buf1, one = 1, *count;
	
	std::random_device rd;
	std::mt19937 e2(rd());
	std::uniform_real_distribution<> dist(0, DISTRIBUTION);
	
	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &count, &win_count);
	foMPI_Win_allocate(sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &throughput, &win_results);
	if(rank == 0){
		*count = 1;
	} else {
		*count = 0;
	}

	/* Warm-up */
	int i;
	for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
		if(std::floor(dist(e2)) < writer_rate){
			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, 0, 0, win_count);
			foMPI_Put(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Win_unlock(0, win_count);
		} else {
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&buf1, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Win_unlock(0, win_count);
		}
	}
assert(buf1);

		
	foMPI_Win_fence(0, win_count);
	HRT_GET_TIMESTAMP(t[0]);
	/* Start test */
	for(i = 0; i < MAX_ITERATIONS; i++){
		if(std::floor(dist(e2)) < writer_rate){
			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, 0, 0, win_count);
			foMPI_Put(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Win_unlock(0, win_count);
		} else {
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&buf1, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Win_unlock(0, win_count);
		}	
	}
			
	HRT_GET_TIMESTAMP(t[1]);
	HRT_GET_ELAPSED_TICKS(t[0], t[1], &ticks_diff);
	time_diff = HRT_GET_USEC(ticks_diff);
	*throughput = MAX_ITERATIONS / (time_diff / USECONDS_FOR_THROUGHPUT);

	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		int j;
		/* Get the count of throughput from each process */
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			double buf;
			foMPI_Get(&buf, 1, MPI_DOUBLE, j, 0, 1, MPI_DOUBLE, win_results);
			foMPI_Win_fence(0, win_results);
			*throughput += buf;
		}
		/* Write result to a file */
		snprintf(&line[0], 256, "%20.4f", *throughput);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_results);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win_count);
	foMPI_Win_free(&win_results);
}

