#include "benchmark.h"
#include <random>

using namespace std;

/*
 * Measuring throughput of foMPI's reader and writer locks with empty CS in one second 
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void rw_fompi_throughput_empty_CS(MPI_Comm comm, MPI_File filehandle) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long ticks_diff;
	double time_diff;
	
	char line[256];
	double *throughput;
	
	foMPI_Win win;
	int comm_size;
	int rank;
	
	std::random_device rd;
	std::mt19937 e2(rd());
	std::uniform_real_distribution<> dist(0, DISTRIBUTION);
	
	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &throughput, &win);

	/* Warm-up */
	int i;
	for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
		if(std::floor(dist(e2)) < writer_rate){
			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, 0, 0, win);
			foMPI_Win_unlock(0, win);
		} else {
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win);
			foMPI_Win_unlock(0, win);
		}
	}
		
	foMPI_Win_fence(0, win);
	HRT_GET_TIMESTAMP(t[0]);
	/* Start test */
	for(i = 0; i < MAX_ITERATIONS; i++){
			if(std::floor(dist(e2)) < writer_rate){
				foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, 0, 0, win);
				foMPI_Win_unlock(0, win);
			} else {
				foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win);
				foMPI_Win_unlock(0, win);
			}

	}
				
	HRT_GET_TIMESTAMP(t[1]);
	HRT_GET_ELAPSED_TICKS(t[0], t[1], &ticks_diff);
	time_diff = HRT_GET_USEC(ticks_diff);
	*throughput = MAX_ITERATIONS / (time_diff / USECONDS_FOR_THROUGHPUT);
	
foMPI_Win_fence(0, win);
	if(rank == 0){
		int j;
		/* Get the count of throughput from each process */
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win);
			double buf;
			foMPI_Get(&buf, 1, MPI_DOUBLE, j, 0, 1, MPI_DOUBLE, win);
			foMPI_Win_fence(0, win);
			*throughput += buf;
		}
		/* Write result to a file */
		snprintf(&line[0], 256, "%20.4f", *throughput);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win);
			foMPI_Win_fence(0, win);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win);
}

