#include <math.h> //log
#include <float.h> //DBL_MAX_MIN
#include <stdlib.h> //qsort
#include "stats.h"


static void get_min_max_variance_standard_deviation_mean(int count, double array[], double *min, double *max , double *average, double *variance, double *std_deviation);
static int cmp(const void *x, const void *y);
static double get_median(int count, double array[]); 
static int count_range(int count, double array[], double min, double max);




void compute_stats(int count, double values[], stat_desc_t *stat){
  
  get_min_max_variance_standard_deviation_mean(count, values , &(stat->min), &(stat->max) , &(stat->avg), &(stat->variance), &(stat->standard_deviation));
  
  stat->median = get_median(count, values);
  
  stat->fail = count_range(count, values , stat->avg - stat->variance * 2, stat->avg + stat->variance * 2);

}


void print_stats(stat_desc_t *stat, char *str, int maxlen){
	snprintf(str, maxlen,"%7.2lf \t&\t %7.2lf \t&\t %7.2lf  \t&\t %7.2lf \n",
									stat->median, /* median RTT time */
									stat->min, /* minimum RTT time */
									stat->avg, /* average */
									stat->standard_deviation
									);
}

void print_header_latency(char *line, int maxlen){
	snprintf(line, maxlen,
					"#%10s \t#\t %10s \t#\t %10s \t#\t %10s &\n",
                    "med (us)", "min (us)", "avg (us)", "std_dev");

}

void print_header_throughput(char *line, int maxlen){
	snprintf(line, maxlen,
					"#%10s \t#\t %10s \t#\t %10s \t#\t %10s &\n",
                    "med (locks/s)", "min (locks/s)", "avg (locks/s)", "std_dev");
}

/* 
* mean = average of the numbers.
* variance = (summation( ( Xi – average of numbers) * ( Xi – average of numbers)) ) / Total no of elements. where i = 1 to N here N is the total no of elements. 
* Standard deviation = Squareroot of the variance.
*/

static void get_min_max_variance_standard_deviation_mean(int count, double array[], double *min, double *max , double *average, double *variance, double *std_deviation){
	*min = DBL_MAX;
	*max = DBL_MIN;
	double sum = 0, sum1 = 0;
	int i;
	/*  Compute the sum of all elements min and max */
    for (i = 0; i < count; i++)
    {
        if(array[i]>*max){
			*max = array[i];	
		}
		if(array[i]<*min){
			*min = array[i];
		}
        sum = sum + array[i];
    }
    *average = sum / (float) count;
    /*  Compute  variance  and standard deviation  */
    for (i = 0; i < count; i++)
    {
        sum1 = sum1 + pow((array[i] - *average), 2);
    }
    *variance = sum1 / (double)count;
    *std_deviation = sqrt(*variance);
}

static int cmp(const void *x, const void *y)
{
  double xx = *(double*)x, yy = *(double*)y;
  if (xx < yy) return -1;
  if (xx > yy) return  1;
  return 0;
}

static double get_median(int count, double array[]) {
  qsort(array, count, sizeof(double), cmp);
  return array[count/2];
}

static int count_range(int count, double array[], double min, double max) {
  int i=0;
  int item= 0;
  for(; i < count; i++) {
    if(array[i] < min || array[i] > max) item++;
  }
  return item;
}

