#include "benchmark.h"

using namespace std;

/*
 * Measuring throughput of the distributed RW lock with one single memory operation in CS witch 4 writers per node
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void threshold_rw_single_memory_read(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long ticks_diff;
	double time_diff;
	
	char line[256];
	double *throughput;
	
	foMPI_Win win_count, win_results;
	int comm_size;
	int rank;
	int64_t buf1, one = 1, *count;

	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &count, &win_count);
	foMPI_Win_allocate(sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &throughput, &win_results);
	if(rank == 0){
		*count = 1;
	} else {
		*count = 0;
	}

	if(rank == 0)
		printf("single read\n----------------------------------------\n");
	
	/* Warm-up */
	int i;
	if(rank % 16 < 4){
		for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
			rw_lock.writer_acquire();
			foMPI_Put(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			rw_lock.writer_release();

		}
	} else {
		for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
			rw_lock.reader_acquire();
			foMPI_Get(&buf1, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			rw_lock.reader_release();
		}
	}
		
	foMPI_Win_fence(0, win_count);
	HRT_GET_TIMESTAMP(t[0]);
	/* Start test */
	if(rank % 16 < 4){
		for (i = 0; i < MAX_ITERATIONS; i++) {
			rw_lock.writer_acquire();
			foMPI_Put(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			rw_lock.writer_release();

		}
	} else {
		for (i = 0; i < MAX_ITERATIONS; i++) {
			rw_lock.reader_acquire();
			foMPI_Get(&buf1, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_flush(0, win_count);
			rw_lock.reader_release();
		}
	}
			
	HRT_GET_TIMESTAMP(t[1]);
	HRT_GET_ELAPSED_TICKS(t[0], t[1], &ticks_diff);
	time_diff = HRT_GET_USEC(ticks_diff);
	*throughput = MAX_ITERATIONS / (time_diff / USECONDS_FOR_THROUGHPUT);

	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		int j;
		/* Get the count of throughput from each process */
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			double buf;
			foMPI_Get(&buf, 1, MPI_DOUBLE, j, 0, 1, MPI_DOUBLE, win_results);
			foMPI_Win_fence(0, win_results);
			*throughput += buf;
		}
		/* Write result to a file */
		snprintf(&line[0], 256, "%20.4f", *throughput);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_results);
			foMPI_Win_fence(0, win_results);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win_count);
	foMPI_Win_free(&win_results);
}

