#include "test_locks.h"

using namespace std;

/* 
 * Test in which half of the threads acquire a writer lock and count up and the other 
 * half reads the value with foMPI-lock
 *
 * Created on: August 25, 2015
 *     Author: Patrick Schmid
 */
void foMPICounter()
{
	int i, world_size, rank;
	MPI_Comm comm;
	int *dummy, count, compare;
	foMPI_Win win_count;
	
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &world_size);
	foMPI_Win_allocate(sizeof(int), sizeof(int), MPI_INFO_NULL, comm, &dummy, &win_count);
	*dummy = 0;
	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		printf("Start foMPI counter test\n-----------------\n\n");
	}
	for(i = 0; i < 10000; i++){
		if((rank % 2 == 1)){
			foMPI_Win_lock(foMPI_LOCK_EXCLUSIVE, 0, 0, win_count);
			foMPI_Get(&count, 1, MPI_INT, 0, 0, 1, MPI_INT, win_count);
			foMPI_Win_flush(0, win_count);
			count++;
			#ifdef PRINT
			printf("%i writer, count: %i\n", rank, count);
			#endif
			foMPI_Put(&count, 1, MPI_INT, 0, 0, 1, MPI_INT, win_count);
			foMPI_Win_unlock(0, win_count);
		} else {
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&count, 1, MPI_INT, 0, 0, 1, MPI_INT, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Get(&compare, 1, MPI_INT, 0, 0, 1, MPI_INT, win_count);
			foMPI_Win_unlock(0, win_count);
			if(compare != count){
				printf("%i Should not be reached! compare: %i, count: %i\n", rank, compare, count);
			}
		}
		foMPI_Win_fence(0, win_count);
	}
	MPI_Barrier(comm);
	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
		foMPI_Get(&count, 1, MPI_INT, 0, 0, 1, MPI_INT, win_count);
		foMPI_Win_unlock(0, win_count);
		if(count == 10000 * ((world_size / 2))){
			printf("PASSED\n");
		} else {
			printf("FAILED, count: %i\n", count);
		}
		printf("\n----------------------------------------------------\n");
	}
}