#include "test_locks.h"

using namespace std;

/* 
 * Test to check whether foMPI_Compare_and_swap and foMPI_Fetch_and_op work properly 
 * with 64 bit int. Only for 2 threads! 
 *
 * Created on: August 25, 2015
 *     Author: Patrick Schmid
 */
void Int64Test(){
	
	#ifdef __DEBUG
	HRT_TIMESTAMP_T t[2];
	unsigned long long time_diff;
	#endif
	
	MPI_Comm comm;
	int rank, i, count = 0;
	int64_t *buf;
	foMPI_Win win;
	
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &buf, &win);
	foMPI_Win_lock_all(0, win);
	
	int64_t buf1, buf2;
	int64_t one = 1;
	int64_t two = 2, null = -1;	
	
	if(rank == 0){
		printf("Start int64 test\n-----------------\n\n");
	}

	for(i = 0; i < 100000; i++)
	{
		*buf = 2;
		foMPI_Win_fence(MPI_INFO_NULL, win);
		if(rank == 0){
			/* Call fetch and op and try to switch what’s there with 1 */
			#ifdef __DEBUG
			HRT_GET_TIMESTAMP(t[0]);
			#endif
		
			foMPI_Fetch_and_op(&one, &buf1, MPI_INT64_T, 0, 0, foMPI_REPLACE, win);
		
			#ifdef __DEBUG
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
	
			printf("time fetch and op: %f\n", HRT_GET_USEC(time_diff));
			#endif
			/* To compare the buffers you have to exchange the one from rank 0 to rank 2 */
			foMPI_Win_fence(MPI_INFO_NULL, win);
			foMPI_Win_fence(MPI_INFO_NULL, win);
			/* Test whether fetch and op and compare & swap got both the same result, if yes that is an error */
			//printf("buf: %" PRId64 ", buf1: %" PRId64 "\n", *buf, buf1);
assert(*buf == 2 || *buf == 1);
assert(buf1 == 2 || buf1 == -1);
			if(*buf == 2 && 2 == buf1){
				count++;
			}
			__sync_synchronize();
		} else {
			/* Call compare & swap and test if there is 2 in the window, if yes change it to -1 */
			#ifdef __DEBUG
			HRT_GET_TIMESTAMP(t[0]);
			#endif
			foMPI_Compare_and_swap(&null, &two, &buf2, MPI_INT64_T, 0, 0, win);
		
			#ifdef __DEBUG
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
	
			printf("time compare and swap: %f\n", HRT_GET_USEC(time_diff));
			#endif
			/* To compare the buffers you have to exchange the one from rank 1 to rank 2 */
			foMPI_Win_fence(MPI_INFO_NULL, win);
			foMPI_Put(&buf2, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win);
			foMPI_Win_fence(MPI_INFO_NULL, win);
		} 
		
	}
	if(rank == 0){
		if(count > 0) {
			printf("FAILED, count: %i\n", count);
		} else {
			printf("PASSED\n");
		}
		printf("\n------------------------------------------------\n");
	}
	foMPI_Win_unlock_all(win);
	foMPI_Win_free(&win);
}
