#include "test_locks.h"
#include <time.h>
#include "Locks.h"

using namespace std;

#ifdef __DEBUG
unsigned long long g_timerfreq;
double g_hrttimer_startvalue;
#endif

int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

/*	
 * Main function for testing
 *
 * Created on: August 25, 2015
 *     Author: Patrick Schmid
 */
int main(int argc, char *argv[]){
	
	int rank;
	MPI_Comm comm;
	MCSLock mcs_lock;
	HMCSLock hmcs_lock;
	RWLock rw_lock;
	
	marker_dist = atoi(argv[1]);
	hmcs_threshold = atoi(argv[2]);
	tree_threshold = atoi(argv[3]);
	root_threshold = atoi(argv[4]);
	reader_threshold = atoi(argv[5]);
	writer_rate = atoi(argv[6]);

	foMPI_Init(&argc, &argv);
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	if(rank == 0){
		printf("START OF TESTING, ");
	}
	#ifdef PRINT
	printf("before init %i\n", rank);
	#endif
	
	if(MCS_RANK_COUNT || MCS_LOCK_A_LOT)
		mcs_lock.initialize();
	if(HMCS_RANK_COUNT || HMCS_LOCK_A_LOT)
		hmcs_lock.initialize();
	if(RW_RANK_COUNT || RW_LOCK_A_LOT || W_LOCK_A_LOT || R_LOCK_A_LOT || RW_COUNTER)
		rw_lock.initialize();
	
	#ifdef __DEBUG
	srand(time(NULL) + rank * 100000);
	HRT_INIT(1, g_timerfreq);
	#endif
	
	if(MCS_RANK_COUNT)
		MCS_rank_count(mcs_lock);
	
	if(MCS_LOCK_A_LOT)
		MCS_lock_a_lot(mcs_lock);
	
	if(HMCS_RANK_COUNT)
		HMCS_rank_count(hmcs_lock);
	
	if(HMCS_LOCK_A_LOT)
		HMCS_lock_a_lot(hmcs_lock);
	
	if(INT64_TEST)
		Int64Test();
	
	if(INT_TEST)
		IntTest();
	
	if(RW_RANK_COUNT)
		RWRankCount(rw_lock);
	
	if(R_LOCK_A_LOT)
		RLockALot(rw_lock);	
	
	if(W_LOCK_A_LOT)
		WLockALot(rw_lock);
	
	if(RW_LOCK_A_LOT)
		RWLockALot(rw_lock);
	
	if(RW_COUNTER)
		RWCounter(rw_lock);
	
	if(foMPI_COUNTER)
		foMPICounter();

	if(RW_RANK_COUNT || RW_LOCK_A_LOT || W_LOCK_A_LOT || R_LOCK_A_LOT || RW_COUNTER)
		rw_lock.finalize();
	if(MCS_RANK_COUNT || MCS_LOCK_A_LOT)
		mcs_lock.finalize();
	if(HMCS_RANK_COUNT || HMCS_LOCK_A_LOT)
		hmcs_lock.finalize();
	//foMPI_Finalize();
	#ifdef PRINT
	printf("after finalize %i\n", rank);
	#endif
	return 0;
}
