#include "test_locks.h"

using namespace std;

/* 
 * Test in which half of the threads acquire a writer lock and count up and the other 
 * half reads the value 
 *
 * Created on: August 25, 2015
 *     Author: Patrick Schmid
 */
void RWCounter(RWLock rw_lock)
{
	int i, world_size, rank;
	MPI_Comm comm;
	int64_t *dummy, count, compare;
	foMPI_Win win_count;
	
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &world_size);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &dummy, &win_count);
	*dummy = 0;
	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		printf("Start rw counter test\n-----------------\n\n");
	}
	for(i = 0; i < 10000; i++){
		if((rank % 2 == 0)){
			rw_lock.writer_acquire();
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&count, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			count++;
			#ifdef PRINT
			printf("%i writer, count: %" PRId64 "\n", rank, count);
			#endif
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Put(&count, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&compare, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			rw_lock.writer_release();
			if(compare != count){
				printf("%i Should not be reached! Writer compare: %" PRId64 ", count: %" PRId64 "\n", rank, compare, count);
			}
		} else {
			rw_lock.reader_acquire();
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&count, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&compare, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			rw_lock.reader_release();
			if(compare != count){
				printf("%i Should not be reached! Reader compare: %" PRId64 ", count: %" PRId64 "\n", rank, compare, count);
			}
		}
	}
	MPI_Barrier(comm);
	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
		foMPI_Get(&count, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
		foMPI_Win_unlock(0, win_count);
		if(count == 10000 * ((world_size / 2) + (world_size % 2))){
			printf("PASSED\n");
		} else {
			printf("FAILED, count: %" PRId64 " \n", count);
		}
		printf("\n----------------------------------------------------\n");
	}
	/* Finalize */
	foMPI_Win_free(&win_count);
}