#include "test_locks.h"
#include <random>

using namespace std;

/* 
 * Test which acquires and releases reader and writer locks very often. Can 
 * adjust rate of writers.
 *
 * Created on: August 25, 2015
 *     Author: Patrick Schmid
 */
void RWLockALot(RWLock rw_lock)
{
	int i;
	int64_t one = 1;
	MPI_Comm comm;
	int rank;
	int64_t *dummy, count;
	foMPI_Win win_count;
	RWLock rw_lock2;
	
	std::random_device rd;
	std::mt19937 e2(rd());
	std::uniform_real_distribution<> dist(0, 100);
	
	rw_lock2.initialize();
	comm = MPI_COMM_WORLD;
	MPI_Comm_rank(comm, &rank);
	foMPI_Win_allocate(sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm, &dummy, &win_count);
	*dummy = 0;
	foMPI_Win_fence(0, win_count);
	if(rank == 0){
		printf("Start rw lock a lot test\n---------------------\n\n");
	}
	for(i = 0; i < 10000; i++){
		if(std::floor(dist(e2)) < writer_rate){
			if(i % 2 == 0){
				rw_lock.writer_acquire();
			} else {
				rw_lock2.writer_acquire();
			}
			foMPI_Get_accumulate(0, 0, MPI_INT64_T, &count, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, foMPI_NO_OP, win_count);
			foMPI_Win_flush(0, win_count);
			foMPI_Accumulate(&one, 1, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, foMPI_SUM, win_count);
			foMPI_Win_flush(0, win_count);
			if(i % 2 == 0){
				rw_lock.writer_release();
			} else {
				rw_lock2.writer_release();
			}
		} else {
			if(i % 2 == 0){
				rw_lock.reader_acquire();
			} else {
				rw_lock2.reader_acquire();
			}
			foMPI_Win_lock(foMPI_LOCK_SHARED, 0, 0, win_count);
			foMPI_Get(&count, 2, MPI_INT64_T, 0, 0, 1, MPI_INT64_T, win_count);
			foMPI_Win_unlock(0, win_count);
			if(i % 2 == 0){
				rw_lock.reader_release();
			} else {
				rw_lock2.reader_release();
			}
		}
	}
	if(rank == 0){
		printf("COUNT: %" PRId64  "\n\n----------------------------------------------------\n", count);
	}
	/* Finalize */
	rw_lock2.finalize();
	foMPI_Win_free(&win_count);
}